import logging
import os
import shutil
from pathlib import Path

from . import __version__ as version
from . import exports as exporter
from . import gene_modules as gmodules
from . import imports as imports
from . import linker as linker
from . import performance as performance
from . import profiles as profiles
from . import utilities as utils
from old_code import red_refinement as red_refiner, viral_clusters as vclusters
from . import species
import pkg_resources
import json

import pickle
import gzip

db_dir = Path(pkg_resources.resource_filename('vcontact3', 'data/'))  # Where to store all the resources
db_info_fp = db_dir / 'db_info.json'


def run_vcontact3(blastp_fp=None, c1_fp=None, diamond_fp=None,
        mmseqs_fp=None, prodigal_fp=None,
        db_domain=None, db_version=None, nucleotides=None,
        gene2genome=None, gene_cluster_mode=None, proteins=None,
        force=False, output_dir=None, refiner=None,
        cluster_id=None, min_shared=None, dist_metric=None,
        exports=None, threads=None, verbose=None, skip_msh=False,
        msh_max_d=0.2, msh_k=16, msh_v=1.0, msh_s=1000, min_af=0.8,
        msh_db=None, extension=None):
    # https://fangpenlin.com/posts/2012/08/26/good-logging-practice-in-python/
    # Logger config
    logger_levels = {
        0: logging.INFO,
        1: logging.WARNING,
        2: logging.ERROR,
        3: logging.DEBUG,
    }
    logging.basicConfig(level=logger_levels[verbose],
                        format='[%(levelname)s] %(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S %p')

    # Create logger
    logger = logging.getLogger(__name__)

    main_intro = f'This is vConTACT3 {version}'
    print(f'\n{main_intro:=^80}\n')

    logger.info('Identifying tool locations...')

    tools = {'MMSeqs2': '',
             'Diamond': '',
             'HMMsearch': '',
             'BLASTP': '',
             'ClusterONE': '',
             'Prodigal': ''}
    # TODO prodigal and pprodigal, or just prodigal
    tools_bin = ['mmseqs', 'diamond', 'blastp', 'cluster_one-1.0.jar', 'prodigal']
    tools_arg = [mmseqs_fp, diamond_fp, blastp_fp, c1_fp, prodigal_fp]

    for tool, tool_fn, tool_arg in zip(tools.keys(), tools_bin, tools_arg):

        if tool_arg:  # User provided
            if os.path.isfile(tool_arg):  # And vConTACT3 can find it
                tools[tool] = tool_arg
            elif os.path.isdir(tool_arg):
                # Search within
                tool_fp = os.path.join(tool_arg, tool_fn)
                if os.path.isfile(tool_fp):  # And vConTACT3 can find it
                    tools[tool] = tool_fp
                else:
                    logger.warning(f'Unable to find {tool} within user-supplied directory, will try other methods')
            else:
                logger.warning(f'Unable to find {tool} using the given user path, will try other methods')

        elif not tool_arg:  # No provided path to tool
            tool_fp = shutil.which(tool_fn)  # Can system find it?
            if not tool_fp:  # is None
                logger.warning(f'Unable to find {tool} in $PATH')
            else:
                tools[tool] = tool_fp

        else:
            logger.warning('Something went wrong with identifying tool paths.')

    logger.info('Identified the following locations for tools. If one is not found, it may not impact the analysis, '
                'unless said tool is required for the analysis (i.e. not finding MCL but needing it to cluster '
                'proteins)\n' + utils.fmt_dict(tools))

    logger.info(f'Identified {threads} CPUs')

    logger.info('Sourcing databases...')

    if db_domain == 'None':
        logger.warning('No reference database selected. vConTACT3 will run in de novo mode.')

    if not db_info_fp.exists():
        logger.error('You need to run prepare_databases --initialize-dbs before proceeding.')
        exit(1)

    with open(db_info_fp, 'r') as db_info_fh:
        db_info = json.load(db_info_fh)

    latest_version = max(db_info["VOGDB"].keys())
    logger.info(f'Using reference database: {db_domain} (version {latest_version})')

    logger.info('Identifying and validating inputs...')

    proteins_fp = None
    nucleotides_fp = None
    gene2genome_fp = None
    if nucleotides:
        if not os.path.exists(nucleotides):
            logger.error('Unable to locate nucleotides file. This could be a simple path issue. Please provide the '
                         'full path to the nucleotide file (often ends in *.fna or *.fasta) if the relative path '
                         'fails.')
        nucleotides_fp = nucleotides
        if proteins and gene2genome:
            logger.warning('Both nucleotide and protein sequences (WITH gene-to-genome file) were provided. Will '
                           'bypass gene prediction. To avoid this warning in the future, either provide just the '
                           'nucleotide sequence file *OR* the protein sequence with the gene-to-genome file.')
            # Don't add nucleotide here to keep nucleotide variable empty
        elif proteins and not gene2genome:
            logger.warning('Both nucleotide and protein sequences (WITHOUT gene-to-genome file) were provided. '
                           'Defaulting to gene prediction. To avoid this warning in the future, either provide just '
                           'the nucleotide sequence *OR* include the gene-to-genome file alongside the protein '
                           'sequences.')
            nucleotides_fp = nucleotides  # Only have nucleotides, must predict
        else:  # Only providing nucleotides
            # Send to gene calling
            nucleotides_fp = nucleotides

    elif proteins:
        # Should catch +nucleotides earlier
        if not proteins.exists():
            logger.error('Unable to locate proteins file. This could be a simple path issue. Please provide the full '
                         'path to the proteins file (often ends in *.faa) if the relative path fails.')
        if gene2genome:
            # If gets here, then should exist
            if gene2genome.exists():
                pass
            else:
                logger.error('Unable to locate gene-to-genome file. This could be a simple path issue. Please provide '
                             'the full path to the gene-to-genome file (should end in *.csv or *.tsv) if the relative '
                             'path fails.')
        elif not gene2genome:
            logger.error('You must provide a gene-to-genome file alongside the proteins file. Otherwise, provide the '
                         'nucleotide sequence and vConTACT3 will predict gene coding regions and create a suitable '
                         'file.')
        else:
            logger.error('Unknown error. Proteins provided. Please send the full log to the vConTACT3 developers.')

    else:
        logger.error('Did not provide a nucleotide sequence file OR a proteins sequence file with gene-to-genome file. '
                     'Analysis cannot continue. If this is a job being re-run (for whatever reason), please provide the'
                     ' same input arguments as before.')

    input_fps = {
        'Nucleotides': nucleotides_fp,
        'Proteins': proteins_fp,
        'Gene-to-genome': gene2genome_fp
    }
    logger.info('Identified the following input files. If you provided a file path and it is not found (i.e. "None") '
                'then check for any warning messages and correct.\n' + utils.fmt_dict(input_fps))

    pre_flight_complete = 'Pre-flight checks complete! Input files have been identified. Beginning analysis'
    logger.info(f'\n{pre_flight_complete:-^80}\n')

    output_dir = utils.check_dir_exists_and_create(output_dir)

    # What kind of inputs do we have?
    if nucleotides_fp:  # TODO Check if exists
        # insert species module code here
        #  return DataFrame of member -> rep -> taxonomy
        #  write nucl sequences of representative seqs

        seqfiles = nucleotides_fp

        d = {'seqfiles': seqfiles, 'no_mash': False, 'max_d': 0.2,
             'out_dir': 'species', 'mash_k': 16, 'mash_v': 1.0,
             'mash_s': 1000, 'min_af': 0.8, 'mash_db': None,
             'extension': None, 'cpus': 1}
        spe_df = species.species(**d)
        # unclassified vOTU seqfile in vOTU_uncl.fa
        #
        logger.info('Predicting genes and generating gene-to-genome mapping file.')
        proteins_fp, gene2genome_fp = imports.call_genes(nucleotides_fp, output_dir, force=force)
    else:  # Need to catch this and 're-direct' any existing protein/gene2genome input to output directory
        logger.info('Identified gene and gene-to-genome mapping file, copying to output directory.')
        proteins_fp = output_dir / proteins.name
        gene2genome_fp = output_dir / gene2genome.name
        shutil.copy2(proteins, proteins_fp)
        shutil.copy2(gene2genome, gene2genome_fp)

    # pprint(db_info)

    profile_fn = proteins_fp.stem
    profile_pkl = output_dir / f'{profile_fn}.{db_domain}.profile.pkl.gz'

    if not profile_pkl.exists():

        profile = profiles.Profiles(threads=threads, method=gene_cluster_mode, proteins=proteins_fp,
                                    gene2genome=gene2genome_fp, ref_db=db_info, ref_version=latest_version,
                                    ref_domain=db_domain,
                                    output_dir=output_dir, evalue=1e-10)

        with gzip.open(profile_pkl, 'wb') as profile_gzip_fh:
            logger.info(f'Saving profiles to {profile_pkl}. If there is a problem during downstream processing, '
                        'vConTACT3 will resume from AT LEAST this point. Enable --force-overwrite if you wish to '
                        'start from scratch.')
            pickle.dump(profile, profile_gzip_fh, pickle.HIGHEST_PROTOCOL)

    else:
        with gzip.open(profile_pkl, 'rb') as profile_gzip_fh:
            profile = pickle.load(profile_gzip_fh)
            logger.info(f'Loaded profiles from {profile_pkl}. This may occur if the previous run abruptly failed or '
                        f'there was some sort of issue. Continuing from this point, or further.')

    # cluster_one_parameters = {
    #     '--min-density': parameters.vc_density,
    #     '--min-size': parameters.vc_size,
    #     '--max-overlap': parameters.vc_overlap,
    #     '--penalty': parameters.vc_penalty,
    #     '--haircut': parameters.vc_haircut,
    #     '--merge-method': parameters.vc_method,
    #     '--similarity': parameters.vc_similarity,
    #     '--seed-method': parameters.vc_seed
    # }
    cluster_one_parameters = {
        '--min-density': 0.3,
        '--min-size': 2,
        '--max-overlap': 0.9,
        '--penalty': 2.0,
        '--haircut': 0.55,
        '--merge-method': 'single',
        '--similarity': 'match',
        '--seed-method': 'nodes'
    }

    mcl_parameters = {
    }

    viral_clstr_pkl = os.path.join(output_dir, f'{profile_fn}.{db_domain}.clusters.pkl.gz')

    if not os.path.exists(viral_clstr_pkl):

        # TODO need refinement step included in viral clusters - "PCs" go in and "VCs" come out
        vc = vclusters.ViralCluster(profile, output_dir=output_dir, tools=tools,
                                    method_opts=cluster_one_parameters)

        with gzip.open(viral_clstr_pkl, 'wb') as viral_clstr_gzip_fh:
            logger.info(f'Saving viral clusters to {viral_clstr_pkl}. If there is a problem during downstream '
                        f'processing, vConTACT3 will resume from AT LEAST this point. Enable --force-overwrite if you '
                        f'wish to start from scratch.')
            pickle.dump(vc, viral_clstr_gzip_fh)

    else:
        with gzip.open(viral_clstr_pkl, 'rb') as viral_clstr_gzip_fh:
            vc = pickle.load(viral_clstr_gzip_fh)
            logger.info(f'Loaded viral clusters from {viral_clstr_pkl}. This may occur if the previous run abruptly '
                        f'failed or there was some sort of issue. Continuing from this point, or further.')

    # TODO VC Refinement - either do though performance (via ICTV taxonomy) or RED
    # In order to calculate RED, need VCs (or rather, "pre"-VCs), but also gene modules present in that pVC cuz
    # RED needs a tree constructed from a concatenated gene tree
    # Can simply use profile's self.genes_overview, which is merged g2g before any modifications

    red_refinement_pkl = os.path.join(output_dir, f'{profile_fn}.{db_domain}.red_refine.pkl.gz')
    refined = False
    if refiner == 'RED':
        # TODO Should refined actually be vc?

        if not os.path.exists(red_refinement_pkl):

            refined = red_refiner.Refiner(profile, vc, ref_db=db_info, ref_version=latest_version,
                                          ref_domain=db_domain, method=gene_cluster_mode, output_dir=output_dir)

            with gzip.open(red_refinement_pkl, 'wb') as red_refinement_gzip_fh:
                logger.info(f'Saving RED-refined VCs to {red_refinement_pkl}. If there is a problem during downstream '
                            f'processing, vConTACT3 will resume from AT LEAST this point. Enable --force-overwrite '
                            f'if you wish to start from scratch.')
                pickle.dump(refined, red_refinement_gzip_fh)

        else:
            with gzip.open(red_refinement_pkl, 'rb') as red_refinement_gzip_fh:
                refined = pickle.load(red_refinement_gzip_fh)
                logger.info(
                    f'Loaded RED-refined VCs from {red_refinement_pkl}. This may occur if the previous run abruptly '
                    f'failed or there was some sort of issue. Continuing from this point, or further.')

    # TODO How to handle RED-refined clusters, i.e. "VCs"

    # So calculate pVCs, then calculate gene modules, then can feed that to performance and exports

    modules_pkl = os.path.join(output_dir, f'{profile_fn}.{db_domain}.modules.pkl.gz')

    if not os.path.exists(modules_pkl):

        modules = gmodules.Module(profile, output_dir=output_dir, tools=tools, method='ClusterONE', method_opts={})

        with gzip.open(modules_pkl, 'wb') as modules_gzip_fh:
            logger.info(f'Saving modules to {modules_pkl}. If there is a problem during downstream processing, '
                        f'vConTACT3 will resume from AT LEAST this point. Enable --force-overwrite if you wish to '
                        f'start from scratch.')
            pickle.dump(modules, modules_gzip_fh)

    else:
        with gzip.open(modules_pkl, 'rb') as modules_gzip_fh:
            modules = pickle.load(modules_gzip_fh)
            logger.info(f'Loaded modules from {modules_pkl}. This may occur if the previous run abruptly failed or '
                        f'there was some sort of issue. Continuing from this point, or further.')

    # Performance of the run
    viral_perf_pkl = os.path.join(output_dir, f'{profile_fn}.{db_domain}.performance.pkl.gz')

    if not os.path.exists(viral_perf_pkl):

        if not os.path.exists(red_refinement_pkl):  # TODO Is this needed, as performance will check what the class is?
            vc_perf = performance.Performance(vc, output_dir, taxon_dist_optimize=True)
        else:
            vc_perf = performance.Performance(refined, output_dir, taxon_dist_optimize=False)

        with gzip.open(viral_perf_pkl, 'wb') as viral_perf_gzip_fh:
            logger.info(f'Saving performances to {viral_perf_pkl}. If there is a problem during downstream processing, '
                        f'vConTACT3 will resume from AT LEAST this point. Enable --force-overwrite if you wish to '
                        f'start from scratch.')
            pickle.dump(vc_perf, viral_perf_gzip_fh)

    else:
        with gzip.open(viral_perf_pkl, 'rb') as viral_perf_gzip_fh:
            vc_perf = pickle.load(viral_perf_gzip_fh)
            logger.info(f'Loaded performances from {viral_perf_pkl}. This may occur if the previous run abruptly '
                        f'failed or there was some sort of issue. Continuing from this point, or further.')

    # Pre-save if issues
    # vc_perf.genomes_overview.to_csv(os.path.join(output_dir, f'{profile_fn}.{db_domain}.genomes_overview.csv'))

    logger.info(f'\n{vc_perf.taxon_perf}')

    for taxon_rank in ['NCBI-order', 'NCBI-family', 'NCBI-subfamily', 'NCBI-genus']:
        print(vc_perf.best_taxon_df[taxon_rank])

    exit()

    vc_perf.taxon_perf.to_csv(os.path.join(output_dir, f'{profile_fn}.{db_domain}.performance.csv'))

    link_pkl = os.path.join(output_dir, f'{profile_fn}.{db_domain}.link.pkl.gz')

    if not os.path.exists(link_pkl):

        link = linker.VCmodules(profile, vc, vc_perf, modules, output_dir)

        with gzip.open(link_pkl, 'wb') as link_gzip_fh:
            logger.info(f'Saving linking info to {link_pkl}. If there is a problem during downstream processing, '
                        f'vConTACT3 will resume from AT LEAST this point. Enable --force-overwrite if you wish to '
                        f'start from scratch.')
            pickle.dump(link, link_gzip_fh)

    else:
        with gzip.open(link_pkl, 'rb') as link_gzip_fh:
            link = pickle.load(link_gzip_fh)
            logger.info(f'Loaded linking info from {link_pkl}. This may occur if the previous run abruptly failed or '
                        'there was some sort of issue. Continuing from this point, or further.')

    export_pkl = os.path.join(output_dir, f'{profile_fn}.{db_domain}.exports.pkl.gz')

    # TODO Handle the updating of export, in case a user wants to add more
    if not os.path.exists(export_pkl):

        export = exporter.Export(vc_perf, profile, exports, output_dir)

        with gzip.open(export_pkl, 'wb') as export_gzip_fh:
            logger.info(f'Saving linking info to {export_pkl}. If there is a problem during downstream processing, '
                        f'vConTACT3 will resume from AT LEAST this point. Enable --force-overwrite if you wish to '
                        f'start from scratch.')
            pickle.dump(export, export_gzip_fh)

    else:
        with gzip.open(export_pkl, 'rb') as export_gzip_fh:
            export = pickle.load(export_gzip_fh)
            logger.info(f'Loaded linking info from {export_pkl}. This may occur if the previous run abruptly failed or '
                        'there was some sort of issue. Continuing from this point, or further.')

    logger.info('Complete!')

    # TODO Rename outputs for consistency - either all use the same basename, or create directory and simple names

    # TODO Predictions of at least genus-level

