"""
Resolve cross-domain connections in the network and assign domains to user sequences
"""

import logging
import math
import sys
import time
import pickle
from collections import Counter
from pathlib import Path
from typing import Literal, get_args
from pprint import pprint
import swifter
import igraph as ig
import warnings

import networkx as nx
import numpy as np
import pandas as pd
import scipy.sparse as sparse
from scipy.cluster.hierarchy import linkage
from scipy.spatial import distance
from tqdm import tqdm
import gzip

from .utilities import read_igraphz, write_igraphz

from . import profiles

warnings.filterwarnings("ignore")

logger = logging.getLogger(__name__)

_TYPES = Literal['VirClust', 'SqRoot', 'Shorter', 'Jaccard']

tqdm.pandas()


class Resolver(object):

    def __init__(self, profile,
                 ref_db: dict,
                 ref_fp: Path,
                 ref_domain: str,
                 metric: str,
                 min_shared_genes: int,
                 threads: int,
                 output_dir: Path,
                 sensitive: bool,
                 ):

        ref_version = ref_fp.stem
        ref_dir = ref_fp.parent

        identities = [0.3, 0.4, 0.5, 0.6, 0.7]
        if isinstance(profile, profiles.Profiles):
            self.genomes_overview = profile.genomes_overview  # Everything

            if not all(str(identity) in profile.identities for identity in identities):
                logger.error(f'The following identities are not present in the overviews: {identities}')
                sys.exit(1)

        else:
            logger.error('This isn\'t a Resolvers object. '
                         'Please send this error message to the author.')
            sys.exit(1)

        self.metric = metric
        self.min_shared_genes = min_shared_genes
        self.threads = threads
        self.output_dir = output_dir
        self.identities = {}
        self.ref_db = ref_db[ref_version]["Domains"][ref_domain]  # TODO Don't know if need entire dict
        # TODO just grab the identities here
        ref_genomes_report_df = pd.read_parquet(ref_dir / self.ref_db['Genomes Report'])

        # EACH identity will have its own profile and network
        # Aggregate composition of each identity's unique genes
        for identity in identities:

            identity = str(identity)
            logger.info(f'Identifying {identity}...')

            if identity not in self.identities:
                self.identities[identity] = {}

            profile_fp = profile.identities[identity]['crosstab_fp']  # Don't need updated_c2g as it --> crosstab
            with gzip.open(profile_fp, 'rb') as profile_fh:
                profile_df = pickle.load(profile_fh)

            curr_identity_genomes = profile.identities[identity]['genomes_overview']

            unique_genes = self.ref_db['identities'][identity]['unique genes']
            distance_matrix_fp = output_dir / f'HMMprofile.{identity}_{metric}_distance_matrix.pkl.gz'
            shared_fp = output_dir / f'HMMprofile.{identity}_{metric}_shared_genes.pkl.gz'
            # TODO Need to expose network export at this point for user
            # ntw_fp = output_dir / f'{name}_{metric}_distance.ntw' if export_ntw else False
            ntw_fp = False
            if all(fp.exists() for fp in [distance_matrix_fp, shared_fp]):

                with gzip.open(distance_matrix_fp, 'rb') as distance_matrix_fh:
                    distance_matrix_df = pickle.load(distance_matrix_fh)

                with gzip.open(shared_fp, 'rb') as shared_fh:
                    shared_df = pickle.load(shared_fh)

            else:
                # Build distance matrix w/ network + shared genes - exclude network, because not needed at this point
                (shared_fp,
                 shared_df,  # Genes shared between genomes
                 ntw_fp,
                 distance_matrix_fp,
                 distance_matrix_df
                 ) = build_dist_matrix(crosstab=profile_df, output_dir=output_dir,
                                       name=profile.identities[identity]['name'], metric=metric, identity=identity,
                                       export_ntw=False)

            # Keep these so they can be reused during GBGA
            self.identities[identity]['shared_fp'] = shared_fp  # "Raw" shared genes
            self.identities[identity]['network_fp'] = ntw_fp
            self.identities[identity]['distance_matrix_fp'] = distance_matrix_fp

            # Before filtering, get a list of all genomes that share at least 1 gene, then filter on those
            # These 1-gene genomes will be filtered out in the next step. But what we don't want to do is lose this
            # information. We'll need to add these back in - but ensure they're not from the same realm
            filtered_network_fp = output_dir / f'HMMprofile.{identity}_{metric}_filtered_ntw.pkl.gz'
            if not filtered_network_fp.exists():

                # Is there a decorated network?
                decorated_network_fp = output_dir / f'HMMprofile.{identity}_{metric}_decorated_ntw.pkl.gz'
                if not decorated_network_fp.exists():

                    composed_network_fp = output_dir / f'HMMprofile.{identity}_{metric}_composed_ntw.pkl.gz'
                    if not composed_network_fp.exists():  # BASE LEVEL
                        composed_ntw = build_igraph_network(distance_matrix_df, shared_df, composed_network_fp)
                    else:
                        logger.info(f'Loading previously composed network from {composed_network_fp}...')
                        load_t = time.time()
                        composed_ntw = read_igraphz(composed_network_fp)
                        logger.info(f'Loaded composed network in {time.time() - load_t} seconds')

                    decorated_ntw = annotate_igraph_network(composed_ntw, ref_genomes_report_df, decorated_network_fp)

                else:
                    logger.info(f'Loading previously decorated network from {decorated_network_fp}...')
                    load_t = time.time()
                    decorated_ntw = read_igraphz(decorated_network_fp)
                    logger.info(f'Loaded decorated network in {time.time() - load_t} seconds')

                logger.info(f'Filtering network for {identity}...')
                g = filter_distances(decorated_ntw, curr_identity_genomes, filtered_network_fp, sensitive, min_shared_genes)

            else:
                logger.info(f'Loading previously filtered network from {filtered_network_fp}...')
                load_t = time.time()
                g = read_igraphz(filtered_network_fp)
                logger.info(f'Loaded network in {time.time() - load_t} seconds')

            self.identities[identity]['filtered_network_fp'] = filtered_network_fp

            logger.info('Preparing for identification of genes shared between references and user sequences')

            # Okay. Round 3 of revising this entire workflow. Already have a high-performance graph, iterable through
            # each component. Since we already have distance matrix, just need to assign realm via refs or unique genes

            component_summary_df = pd.DataFrame(
                index=[i for i, n in enumerate(set(g.vs['component']))],
                columns=['References', 'User', 'Realms'])

            # Set user and reference sequences
            user_seqs = curr_identity_genomes[~(curr_identity_genomes['Reference'])].index.tolist()
            ref_seqs = curr_identity_genomes[curr_identity_genomes['Reference']].index.tolist()

            g.vs.select(name_in=ref_seqs)['reference'] = True
            g.vs.select(name_in=user_seqs)['reference'] = False

            # THIS ENTIRE BLOCK, UNTIL CONNECTED COMPONENTS, is simply assigning realm PCs to each member

            # Too long to .loc members and see if they share unique genes with references
            ref_gene_counts = {}
            for realm, unique_gene_fp in unique_genes.items():
                ref_profile_df = pd.read_parquet(ref_dir / unique_gene_fp)

                logger.info(f'Identifying genes shared within references of realm: {realm}')

                if int(ref_version) > 220:

                    # TODO: Function of class or utility?
                    # Should this just go straight to scipy?
                    row_dict = {k: n for n, k in enumerate(ref_profile_df.row.unique())}
                    col_dict = {k: n for n, k in enumerate(ref_profile_df.col.unique())}

                    ref_profile_df['col'] = ref_profile_df['col'].map(col_dict)
                    ref_profile_df['row'] = ref_profile_df['row'].map(row_dict)

                    ref_coo_df = sparse.coo_matrix((ref_profile_df['data'], (ref_profile_df['row'], ref_profile_df['col'])))

                    pcs_for_genome = {row: [] for row in row_dict.keys()}
                    for i, j, _ in zip(ref_coo_df.row, ref_coo_df.col, ref_coo_df.data):
                        genome = list(row_dict.keys())[list(row_dict.values()).index(i)]  # Find genome name by index
                        pc = list(col_dict.keys())[list(col_dict.values()).index(j)]  # Find pc name by index
                        pcs_for_genome[genome].append(pc)

                    for genome, genome_pcs in tqdm(pcs_for_genome.items(),
                                                   total=len(pcs_for_genome),
                                                   unit='genome PCs'):
                        try:
                            node = g.vs.find(name=genome)

                            g.vs[node.index][f'{realm}_PCs'] = genome_pcs

                            if realm not in ref_gene_counts:
                                ref_gene_counts[realm] = Counter(genome_pcs)
                            else:
                                ref_gene_counts[realm].update(genome_pcs)

                        except ValueError:
                            continue

                else:

                    for genome, pcs in tqdm(ref_profile_df.iterrows(),
                                            total=len(ref_profile_df),
                                            unit=' genome PCs'):
                        pcs = pcs[pcs >= 1]

                        try:
                            node = g.vs.find(name=genome)

                            if len(pcs) > 0:
                                g.vs[node.index][f'{realm}_PCs'] = pcs.index.tolist()

                                if realm not in ref_gene_counts:
                                    ref_gene_counts[realm] = Counter(pcs.index.tolist())
                                else:
                                    ref_gene_counts[realm].update(pcs.index.tolist())

                        except ValueError:  # Not in graph
                            continue

            ref_unique_gene_counts = {}

            logger.info(f'Ensuring no PCs (in reference genomes) are shared between realms...')
            shared_keys = set(ref_gene_counts.keys()) & set(unique_genes.keys())
            for realm in shared_keys:
                curr_realm = ref_gene_counts[realm]

                for comp_realm in shared_keys:
                    if realm != comp_realm:  # Self
                        comp_realm_counts = ref_gene_counts[comp_realm]
                        shared_keys = set(curr_realm.keys()) & set(comp_realm_counts.keys())

                        for shared_key in shared_keys:
                            del curr_realm[shared_key]

                ref_unique_gene_counts[realm] = curr_realm

            for realm in shared_keys:
                ref_unique_gene_counts[realm] = {k: v for k, v in ref_unique_gene_counts[realm].items() if v >= 3}

            # Can't iterrows if too many user sequences, as pd will copy and make int64 and consume 100s of GBs (works
            # superfast if not many user sequences). And np.vectorize, pd.apply, pd.swifter.apply are all very slow

            logger.info(f'Identifying genes shared within user sequences (references still counted)...')

            profile_coo = profile_df.sparse.to_coo()
            profile_rows = profile_df.index.tolist()
            profile_cols = profile_df.columns.tolist()

            # Filter profile HERE to only include user sequences?
            # Speed loss by filtering lists exceeds speed gained in iteration

            for i, j, v in tqdm(zip(profile_coo.row, profile_coo.col, profile_coo.data), total=len(profile_coo.data),
                                unit=' genome-PCs'):
                genome = profile_rows[i]
                pc = profile_cols[j]

                if genome in user_seqs:

                    try:
                        node = g.vs.find(name=genome)

                        try:
                            g.vs[node.index][f'user_PCs'].extend([pc])

                        except KeyError:
                            g.vs[node.index][f'user_PCs'] = [pc]

                        except AttributeError:
                            g.vs[node.index][f'user_PCs'] = [pc]

                    except ValueError:
                        continue

            logger.info(f'Identifying shared PCs between user and references')
            for user_genome in user_seqs:

                try:
                    node = g.vs.find(name=user_genome)

                    if 'user_PCs' in node.attributes():
                        pcs = g.vs[node.index][f'user_PCs']
                    else:
                        continue

                except ValueError:
                    continue

                # Do we want to do this for every single genome? >> 4K genomes/sec
                for realm, realm_counts in ref_unique_gene_counts.items():
                    shared_genes = set(pcs) & set(realm_counts.keys())

                    if shared_genes:
                        g.vs[node.index][f'{realm}_shared_PCs'] = shared_genes

            logger.info('Iterating through network components and identifying realms')

            # Go through component to assign
            # This block breaks down refs/users and realm "support"
            # for c in g.vs['component']:  # TODO Switch to network-based approach
            for n, c in enumerate(g.connected_components(mode='weak')):  # ...
                # Assign all the attributes
                c_members = g.vs[c]["name"]  # Human readable

                unique_gene_realms = Counter()  # Only if user genomes exist and are shared with unique genes of refs
                for realm in ref_unique_gene_counts.keys():
                    if f'{realm}_shared_PCs' in g.vs[c].attributes():  # Is this efficient? Canonical?
                        user_shared_with_refs = [s for s in g.vs[c][f'{realm}_shared_PCs'] if s]
                        unique_gene_realms.update([realm] * len(user_shared_with_refs))

                        component_summary_df.at[n, f'{realm} (unique genes)'] = len(user_shared_with_refs)

                # Set how many reference and user sequences are present in component
                # TODO g defines the refs in this component, this set seems like its redundant or can be simplified
                c_refs = curr_identity_genomes.loc[curr_identity_genomes.genome_id.isin(c_members) &
                                                   curr_identity_genomes['Reference']].index.tolist()
                c_user = curr_identity_genomes.loc[curr_identity_genomes.genome_id.isin(c_members) &
                                                   ~(curr_identity_genomes['Reference'])].index.tolist()

                component_summary_df.at[n, 'References'] = len(c_refs)
                component_summary_df.at[n, 'User'] = len(c_user)

                curr_identity_genomes.loc[c_members, 'component'] = int(n)

                # Only references have realms
                realm_counts = Counter([r for r in g.vs[c]['realm'] if r])  # Some refs don't have an assigned realm...
                realm_keys = list(realm_counts.keys())

                # At this point, only use 'Realms' (if exists) and realm prediction... which swiftly gets overridden...

                if len(realm_keys) == 1:  # Only 1 reference realm

                    component_summary_df.at[n, 'Realms'] = realm_keys[0]
                    curr_identity_genomes.loc[c_members, 'realm prediction'] = realm_keys[0]  # priority

                    # At this point, only need to adjust support, as references supersedes
                    # Actually, support is never used. And if needed, can be determined from igraph directly or
                    # calculated on the fly

                elif len(realm_keys) > 1:  # More than 1 reference realm

                    ranked_refs = [k for k, v in realm_counts.most_common()]
                    component_summary_df.at[n, 'Realms'] = '|'.join(ranked_refs)

                    if not unique_gene_realms:  # No unique genes to help
                        curr_identity_genomes.loc[c_members, 'realm prediction'] = '|'.join(ranked_refs)

                    else:
                        unique_gene_realms_keys = list(unique_gene_realms.keys())

                        if len(unique_gene_realms_keys) == 1:  # Only 1 unique

                            if set(unique_gene_realms_keys).issubset(set(realm_keys)):  # And it's a subset
                                curr_identity_genomes.loc[c_members, 'realm prediction'] = '|'.join(ranked_refs)

                            else:  # It disagrees?!
                                ranked_refs = ranked_refs + list(unique_gene_realms_keys)
                                curr_identity_genomes.loc[c_members, 'realm prediction'] = '|'.join(ranked_refs)

                        else:  # Multiple references and multiple unique genes

                            if set(unique_gene_realms_keys).issubset(set(realm_keys)):
                                curr_identity_genomes.loc[c_members, 'realm prediction'] = '|'.join(ranked_refs)

                            else:  # And everyone disagrees

                                ranked_unique_genes = [k for k, v in unique_gene_realms.most_common()]

                                preds = set(ranked_refs) | set(ranked_unique_genes)
                                curr_identity_genomes.loc[c_members, 'realm prediction'] = '|'.join(preds)

                elif not realm_keys:  # No references, or no references with realms

                    # No assign realm as no refs to assign

                    if not unique_gene_realms:  # If unique genes can't help, no choice but to assume to prediction
                        curr_identity_genomes.loc[c_members, 'realm prediction'] = 'No prediction'

                    else:
                        unique_gene_realms_keys = list(unique_gene_realms.keys())

                        if len(unique_gene_realms_keys) == 1:  # Only 1 unique
                            curr_identity_genomes.loc[c_members, 'realm prediction'] = unique_gene_realms_keys[0]

                        else:  # There's more than 1, cuz can't be 0
                            ranked_unique_genes = [k for k, v in unique_gene_realms.most_common()]
                            curr_identity_genomes.loc[c_members, 'realm prediction'] = '|'.join(ranked_unique_genes)

                else:
                    logger.warning(f'Unable to determine realm for component {n} with genomes: {c_members}')

            self.identities[identity]['genomes_overview'] = curr_identity_genomes
            self.identities[identity]['components'] = component_summary_df

            # Maintain to keep track of isolated and edgeless
            self.identities[identity]['edgeless'] = g.vs.select(lambda v: v['edgeless'] == True)['name']
            self.identities[identity]['isolated'] = g.vs.select(lambda v: v['isolated'] == True)['name']

            g_fp = output_dir / f'HMMprofile.{identity}_{metric}_igraph.pkl.gz'
            self.identities[identity]['igraph-network'] = write_igraphz(g, g_fp)

            logger.info(f'Completed identity {identity}')

            # Resolver does not need to hierarchical clustering anything, it just needs to assign

    def __repr__(self):
        mes = (
            "Clustering object {name}, consisting of the following:"
        )
        return mes.format(
            name=self.name
        )


def build_dist_matrix(crosstab: pd.DataFrame, output_dir: Path, name: str, identity: float, metric: _TYPES = 'SqRoot',
                      export_ntw: bool = False):

    """

    Builds a distance matrix using the HMM/PC Profile X Genome matrix

    :param crosstab:
    :param output_dir: Output directory
    :param identity:
    :param metric: Distance metric to apply to the profile
    :param export_ntw: Whether to export the edgelist network
    :param name:
    """

    # https://stackoverflow.com/questions/9522877/pythonic-way-to-have-a-choice-of-2-3-options-as-an-argument-to-a-function
    options = get_args(_TYPES)
    assert metric in options, f"'{metric}' is not in {options}"

    # Need this information to re-map numpy arrays back to human-readable
    module_pos = crosstab.columns.values
    genome_pos = crosstab.index.values

    logger.info(f'Building distance matrix from PC profile')
    # matrix = sparse.csr_matrix(crosstab.to_numpy(dtype=limit))
    matrix = crosstab.sparse.to_coo().tocsr()
    logger.info(f'{matrix.data.nbytes / (1024 * 1024)} megabytes consumed')
    commons_hmms = matrix.dot(matrix.T)  # Can this be cast as int8 too?
    logger.info(f'Finished matrix, proceeding to conversion')

    if np.any(commons_hmms.toarray() < 0):  # Hit if dtype is too large
        logger.error(f'Negative values in common PCs, cannot continue')
        exit(1)
    # TODO Consider avoiding check and also toarray()

    genomes, unique_hmms = matrix.shape  # Number of genomes, number of unique HMMs

    per_genome_HMMs = matrix.sum(1)
    per_genome_HMMs = per_genome_HMMs.A1  # HMMs for each genome, flattened (position based!)

    total_common_hmms = float(commons_hmms.getnnz())

    # Save in order to filter/test for minimum shared genes
    shared_HMMs_fp = output_dir / f'HMMprofile.{identity}_{metric}_shared_genes.pkl.gz'
    shared_HMMs_df = pd.DataFrame.sparse.from_spmatrix(data=commons_hmms,
                                                       index=crosstab.index.tolist(),
                                                       columns=crosstab.index.tolist())  # dtype mirrors original

    # Checked, don't try to downcast int
    with gzip.open(shared_HMMs_fp, 'wb') as shared_HMMs_fh:
        pickle.dump(shared_HMMs_df, shared_HMMs_fh)

    start = time.time()
    logger.info(f"Building genome network with {genomes} contigs and {unique_hmms} HMMs")

    commons_hmms = commons_hmms.todok()
    shared_genomes = [pair for pair in zip(*commons_hmms.nonzero())]  # Tuples list w/ contigs share PCs

    process_list = []
    for common_contig in shared_genomes:
        process_list.append([common_contig[0], common_contig[1], commons_hmms, per_genome_HMMs])

    distances = []
    if metric == 'VirClust':
        for (cA, cB, c_hmm, per_g_hmm) in tqdm(process_list, total=len(shared_genomes)):
            virclust = virclust_dist(cA, cB, c_hmm, per_g_hmm)
            distances.append([cA, cB, virclust])

    elif metric == 'SqRoot':

        for (cA, cB, c_hmm, per_g_hmm) in tqdm(process_list, total=len(shared_genomes)):
            sqroot = sqroot_dist(cA, cB, c_hmm, per_g_hmm)
            distances.append([cA, cB, sqroot])

    elif metric == 'Shorter':

        for (cA, cB, c_hmm, per_g_hmm) in tqdm(process_list, total=len(shared_genomes)):
            shorter = shorter_dist(cA, cB, c_hmm, per_g_hmm)
            distances.append([cA, cB, shorter])

    elif metric == 'Jaccard':

        for (cA, cB, c_hmm, per_g_hmm) in tqdm(process_list, total=len(shared_genomes)):
            jaccard = jaccard_dist(cA, cB, c_hmm, per_g_hmm)
            distances.append([cA, cB, jaccard])

    elif metric == 'Euclidean':
        logger.error(f"Metric '{metric}' is not implemented yet")
        exit(1)

    # print(f'Chunksize: {math.ceil(len(shared_genomes) / threads)}')
    # significances = process_map(alt_dist, process_list,
    #                             max_workers=threads,
    #                             chunksize=math.ceil(len(shared_genomes) / threads))

    if len(distances) == 0:
        logger.error(
            "No edges are found in genome_network;"
            " This is not supposed to happen since"
            " Ref genomes will have edges.."
        )
        sys.exit(1)
    else:
        rows, cols, data = zip(*distances)

    # More efficient to use coo_matrix
    S = sparse.coo_matrix((data, (rows, cols)), shape=(genomes, genomes), dtype='float32')
    S = S.tocsr()
    # S += S.T  # Symmetry
    # S memory size is tiny - at the moment - no need to regulate dtype

    logger.info(f"Genome-similarity/distance network : {genomes} genomes, {S.getnnz()} edges")

    end = time.time()
    t = end - start
    logger.info(f"Genome network building time: {t:.3f}s")

    if export_ntw:

        ntw_fp = output_dir / f'{name}_{metric}_distance.ntw'

        if not ntw_fp.exists():
            logger.info(f"Writing genome edges to {ntw_fp}. This file can be used by numerous 3rd party programs when "
                        f"imported as an edgelist.")

            with open(ntw_fp, "wt") as out_fh:
                pairs = zip(*S.nonzero())
                # dok is efficient at accessing elements
                dok_matrix = sparse.dok_matrix(S)
                for r, c in tqdm(pairs, unit=" genome edges written"):
                    # Need to keep track of which nodes go into network for outlier analysis

                    out_fh.write(
                        " ".join(
                            [str(x) for x in (genome_pos[r], genome_pos[c], dok_matrix[r, c],)]
                        )
                    )
                    out_fh.write("\n")

        else:
            logger.warning(f"Identified an existing network file at {ntw_fp}, reusing...")

    else:
        logger.info(f"Not exporting genome network...")
        ntw_fp = None

    distance_matrix_fp = output_dir / f'HMMprofile.{identity}_{metric}_distance_matrix.pkl.gz'
    if not distance_matrix_fp.exists():

        distance_df = pd.DataFrame.sparse.from_spmatrix(data=S, columns=genome_pos,
                                                        index=genome_pos)  # 23206400 bytes (23 MB)

        with gzip.open(distance_matrix_fp, 'wb') as distance_matrix_fh:
            pickle.dump(distance_df, distance_matrix_fh)

    else:
        logger.warning(f"Identified an existing distance matrix file at {distance_matrix_fp}, reusing...")
        with gzip.open(distance_matrix_fp, 'rb') as distance_matrix_fh:
            distance_df = pickle.load(distance_matrix_fh)

    return shared_HMMs_fp, shared_HMMs_df, ntw_fp, distance_matrix_fp, distance_df


def virclust_dist(genomeA, genomeB, shared_HMMs: np.array, per_genome_HMM: np.array):

    HMMs_A, HMMs_B = per_genome_HMM[genomeA], per_genome_HMM[genomeB]
    shared_HMM = shared_HMMs[genomeA, genomeB]

    dist = 1 - ((2 * shared_HMM) / (HMMs_A + HMMs_B))

    return dist


def sqroot_dist(genomeA, genomeB, shared_HMMs: np.array, per_genome_HMM: np.array):

    HMMs_A, HMMs_B = per_genome_HMM[genomeA], per_genome_HMM[genomeB]
    shared_HMM = shared_HMMs[genomeA, genomeB]

    dist = 1 - (shared_HMM / math.sqrt(HMMs_A * HMMs_B))

    return dist


def shorter_dist(genomeA, genomeB, shared_HMMs: np.array, per_genome_HMM: np.array):

    HMMs_A, HMMs_B = per_genome_HMM[genomeA], per_genome_HMM[genomeB]
    shared_HMM = shared_HMMs[genomeA, genomeB]

    try:
        dist = 1 - (shared_HMM / np.amin([HMMs_A, HMMs_B]))
    except np.AxisError as e:
        print(f'Numpy error: {e}\nShared HMMs: {shared_HMM}\nHMMs_A: {HMMs_A}\nHMMs_B: {HMMs_B}')
        exit(0)

    return dist


def jaccard_dist(genomeA, genomeB, shared_HMMs: np.array, per_genome_HMM: np.array):

    HMMs_A, HMMs_B = per_genome_HMM[genomeA], per_genome_HMM[genomeB]
    shared_HMM = shared_HMMs[genomeA, genomeB]

    dist = 1 - float(shared_HMM / ((HMMs_A + HMMs_B) - shared_HMM))

    return dist


def build_igraph_network(distance_matrix: pd.DataFrame, shared_matrix: pd.DataFrame, out_fp: Path):

    """
    :param distance_matrix:
    :param shared_matrix:
    :param out_fp:
    """

    # Distance matrices maintain dtypes until "conversion" to networkx
    distance_S = distance_matrix.sparse.to_coo()
    shared_S = shared_matrix.sparse.to_coo()

    logger.info(f'Building distance and shared gene networks')
    distG = nx.from_scipy_sparse_array(distance_S, create_using=nx.Graph, parallel_edges=False, edge_attribute='distance')
    sharedG = nx.from_scipy_sparse_array(shared_S, create_using=nx.Graph, parallel_edges=False, edge_attribute='shared')
    logger.info(f'Finished building...')

    # Due to the nature of sparse matrices, and the fact they can self-match, all genomes should be in the network!
    logger.info(f'Building mapping between genomes and ids')
    mapping = dict(zip(range(len(distance_matrix.index.tolist())), distance_matrix.index.tolist()))  # genomes x genomes
    distG = nx.relabel_nodes(distG, mapping)  # Or else it's all ints
    sharedG = nx.relabel_nodes(sharedG, mapping)  # Or else it's all ints

    # Singletons cannot exist from a scipy sparse matrix

    # Remove self-loops from both
    logger.info(f'Removing self-loops from both distance and shared gene networks')
    distG.remove_edges_from(list(nx.selfloop_edges(distG)))  # And this would be dist = 0.0
    sharedG.remove_edges_from(list(nx.selfloop_edges(sharedG)))  # Here there will be max shared genes

    # Remove the newfound isolates
    logger.info(f'Removing isolated nodes from distance and shared gene networks')
    # TODO Keep track of isolates - these are the first
    isolates = list(nx.isolates(distG))
    # distG.remove_nodes_from(isolates)
    # sharedG.remove_nodes_from(list(nx.isolates(sharedG)))
    # TODO NOT removing isolates avoids separating out singletons to iterate through, and introduces the overhead of
    # needing to keep track of the isolates in a separate dataframe or list, BUT dramatically increases processing time

    logger.info(f'Merging shared gene and distance networkx')
    G = nx.compose(distG, sharedG)

    # At this point, a lot of memory is allocated
    del distG
    del sharedG

    logger.info(f'Loading into iGraph')
    g = ig.Graph.from_networkx(G)

    g.vs['name'] = g.vs['_nx_name']
    edgeless_vertices = g.vs.select(name_in=isolates)
    edgeless_indices = [v.index for v in edgeless_vertices]
    g.vs[edgeless_indices]['edgeless'] = True

    del G

    write_igraphz(g, out_fp)

    return g


def annotate_igraph_network(network: ig.Graph, ref_genomes_df: pd.DataFrame, out_fp: Path):

    # TODO GenomeName does not exist for user sequences

    network.vs['name'] = network.vs['_nx_name']

    logger.info(f'Decorating network with reference information wherever possible...')
    annnotation_cols = ['GenomeName', 'RefSeqID', 'realm', 'superkingdom', 'kingdom', 'phylum', 'class',
                        'order', 'family', 'subfamily', 'genus']
    reference_df = ref_genomes_df[annnotation_cols]

    for ref, node_values in tqdm(reference_df.iterrows(), total=len(reference_df), unit=' references'):

        refid = node_values['RefSeqID']
        node_values = node_values.to_dict()

        try:
            node = network.vs.find(name=refid)

            for k, v in node_values.items():
                if not pd.isnull(v):
                    node[k] = v

        except ValueError:  # Not all references are in the network
            continue

    write_igraphz(network, out_fp)

    return network


def filter_distances(ag: ig.Graph, modifying_df, out_fp: Path, enhanced_sensitivity: bool, threshold: int = 3):

    # We want to keep distances, as they're useful to organize the upper ranks, but they also
    """

    :param ag: Annotated graph in igraph format
    :param modifying_df:
    :param out_fp:
    :param threshold: Minimum number of shared genes to be considered
    """

    logger.info(f'Filtering edges at {threshold} to reduce cross-realm connections')
    filtered_edges = [(e, e.source, e.target, e['distance'], e['shared']) for e in ag.es if e['shared'] < threshold]
    # This is the FIRST time that we filter distances. Benchmarks used ">= shared_min" so keeping consistency
    ag.delete_edges([e[0] for e in filtered_edges])
    if len(filtered_edges) == 0:
        logger.warning('No edges were filtered')
    logger.info(f'Removed {len(filtered_edges)} edges')

    logger.info('Identifying further cross-realm components potentially introduced by density-dependent...')
    connected_components = ag.connected_components(mode='weak')
    ag.vs['rawComponent'] = connected_components.membership  # Assign everyone a membership

    ugraphs = []

    filtered_pairs = []
    multi_realm_cc = 0
    for n, cc_members in enumerate(connected_components):

        realm_counts = Counter(r for r in ag.vs[cc_members]['realm'] if r)

        subgraph = ag.subgraph(cc_members)

        if not realm_counts:  # Entirely unknown?
            ugraphs.append(subgraph)

        elif len(realm_counts) == 1:  # Only one realm
            ugraphs.append(subgraph)

        elif len(realm_counts) >= 2:  # Will 3 create situations where some parts more affected than others

            multi_realm_cc += 1

            if subgraph.ecount() == 0:
                logger.warning(f'rawComponent {n} has 0 edges. Graph is inconsistent!')

            i_thres = 3

            while True:

                # These edges are only valid WITHIN this component. When they get unioned, their index will change
                filtered_pair = [
                    (e, subgraph.vs[e.source]['name'], subgraph.vs[e.target]['name'], e['distance'], e['shared']) for e in
                    subgraph.es if e['shared'] <= i_thres]
                filtered_pairs.extend(filtered_pair)

                # Delete edges...
                subgraph.delete_edges([e[0] for e in filtered_pair])

                multi_realms = 0
                for pruned_mem in subgraph.connected_components(mode='weak'):
                    pruned_counts = Counter(r for r in subgraph.vs[pruned_mem]['realm'] if r)

                    if len(pruned_counts) > 1:  # Can't be less than
                        multi_realms += 1

                if multi_realms == 0:
                    break

                # else, there ARE multiple realms and need to remove even more
                i_thres += 1

            ugraphs.append(subgraph)

    ugraph = ig.disjoint_union(ugraphs)

    updated_connected_components = ugraph.connected_components(mode='weak')
    ugraph.vs['preFixComponent'] = updated_connected_components.membership  # Assign everyone a membership

    def validator(graph: ig.Graph):

        for i, i_members in enumerate(graph.connected_components(mode='weak')):

            i_realms = Counter(r for r in graph.vs[i_members]['realm'] if r)
            if len(i_realms) >= 2:
                logger.warning(f'preFixComponent {i} has {len(i_realms)} realms')

    validator(ugraph)

    # Repair edges that are direct references
    logger.info('Repairing edges removed during multi-realm component analysis')
    edges_to_add = []
    edges_to_resolve = []
    for iv_, u, v, w, s in tqdm(filtered_pairs, total=len(filtered_pairs), unit=' filtered edges'):

        u_r, v_r = ugraph.vs.find(name=u)['realm'], ugraph.vs.find(name=v)['realm']

        if pd.isnull(u_r) or pd.isnull(v_r):  # If either is unknown, we're not checking for this yet
            edges_to_resolve.append((u, v, w, s))

        elif u_r == v_r:  # Don't matter what strength, they're references to each other
            edges_to_add.append((u, v, w, s))

        else:
            edges_to_resolve.append((u, v, w, s))  # Could just {}.subset{}, but would require adding another to list

    logger.info(f'Recovering {len(edges_to_add)} realm-realm edges lost during filtering...')
    es_end = ugraph.ecount()
    ugraph.add_edges([(e[0], e[1]) for e in edges_to_add])
    ugraph.es[es_end:]['distance'] = [e[2] for e in edges_to_add]
    ugraph.es[es_end:]['shared'] = [e[3] for e in edges_to_add]

    def iterative_repair(graph: ig.Graph, to_filter_edges: list):

        """
        Reset each time, unless vertices are deleted, should be maintained

        """

        component_to_realm = {}  # Set up dict connecting component to realm
        member_to_component = {}  # Set up dict connecting member to component
        for i_component, component in enumerate(graph.connected_components(mode='weak')):
            revised_counts = Counter([r for r in graph.vs[component]['realm'] if r])

            if len(revised_counts) > 1:
                print('Warning: Identified multiple realms within subcomponent: ', revised_counts)
                exit(1)

            if revised_counts:
                component_to_realm[i_component] = revised_counts.most_common(1)[0][0]
            else:
                component_to_realm[i_component] = np.nan

            for member in component:
                member_name = graph.vs[member]['name']
                member_to_component[member_name] = i_component

        to_add_edges = []
        for u, v, w, s in tqdm(to_filter_edges[::-1], total=len(to_filter_edges), unit=' iter edges'):

            # Get components
            u_c = member_to_component[u]
            v_c = member_to_component[v]
            # Get component realms
            u_c_realm = component_to_realm[u_c]
            v_c_realm = component_to_realm[v_c]

            if u_c_realm == v_c_realm:  # Two components, now merged, have "added in" references
                to_add_edges.append((u, v, w, s))

            elif pd.notnull(u_c_realm) and pd.notnull(v_c_realm):
                if u_c_realm != v_c_realm:  # Don't want to merge components if each are of different realms
                    continue
                else:
                    logger.warning(f'Unable to identify realms. Both exist, but are neither the same nor different')
                    print(u, u_c, u_c_realm, v, v_c, v_c_realm)

            elif pd.notnull(u_c_realm) or pd.notnull(v_c_realm):  # Not both
                # It's possible for edges DURING ITERATION to connect ref-unknown and that same unknown-diff ref
                # Let the specialized function handle these cases
                continue

            elif pd.isnull(u_c_realm) and pd.isnull(v_c_realm):
                continue  # Don't want to merge two subcomponents that have no realm

            else:
                logger.warning(f'Unable to identify {u} and {v}, with {u_c} and {v_c}, and {u_c_realm} and {v_c_realm}')
                continue

        if not to_add_edges:  # Continue until no more components can be merged
            return to_filter_edges

        else:
            logger.info(f'Repairing {len(to_add_edges)} edges from filtering...')
            ee_end = graph.ecount()
            graph.add_edges([(e[0], e[1]) for e in to_add_edges])
            graph.es[ee_end:]['distance'] = [e[2] for e in to_add_edges]
            graph.es[ee_end:]['shared'] = [e[3] for e in to_add_edges]

            return list(set(to_filter_edges) - set(to_add_edges))

    # Round 1 (after reference-reference)
    iteration = 0
    while True:

        logger.info(f'Processing repair iteration {iteration}, starting with {len(edges_to_resolve)} edges')
        remaining_edges = iterative_repair(ugraph, edges_to_resolve)  # How many remains

        if len(remaining_edges) == len(edges_to_resolve):  # Hasn't decreased
            break

        iteration += 1  # Update remaining to include how many the current iteration has remaining
        edges_to_resolve = remaining_edges

    validator(ugraph)

    # The question is, how to repair components that could connect to other components that are connected to diff realms?
    # Problem is that DURING ITERATION, could connect two different realm components through an unknown intermediate
    # So, basically, repeat iterative_repair, but add edges one-by-one

    logger.debug(f'Writing temporary output graph')
    write_igraphz(ugraph, out_fp.parent / f'{out_fp.name.replace(".pkl.gz", "_temp.pkl.gz")}')

    edges_to_resolve.sort(key=lambda x: x[2])

    def iterative_repair_2nd(graph: ig.Graph, to_filter_edges: list):

        """
        A variant of iterative_repair, but instead of adding edges in batch (i.e. the faster and "how to do it" way),
        this repair approach adds edges sequentially, and then updates the mapping dictionaries. This is to control for
        the fact that adding an edge between two components can/will "guilt" a component into assuming the realm of its
        larger merging component partner. But if that guilted component is also connected to another component with a
        different realm... then that component would suddenly now connect two differing realms. While this does -
        biologically - occur, it complicates assignments.

        :param graph: iGraph graph that must have the "name" attribute and at least one "realm" attribute
        :param to_filter_edges: List of edges (of u, v, s, w) to be added to the graph
        """

        # Set up dict connecting member to component
        # Set up dict connecting component to realm
        component_to_realm = {}  # component_num: realm
        member_to_component = {}  # name: component_num
        component_counts = Counter()  # component_num: count
        for component_i, component in enumerate(graph.connected_components(mode='weak')):
            revised_counts = Counter([r for r in graph.vs[component]['realm'] if r])

            if len(revised_counts) > 1:
                logger.warning('Identified multiple realms of subcomponent: ', revised_counts, ' graph is inconsistent')
                exit(1)

            if revised_counts:
                component_to_realm[component_i] = revised_counts.most_common(1)[0][0]
            else:
                component_to_realm[component_i] = np.nan

            for mem in component:
                mem_name = graph.vs[mem]['name']
                member_to_component[mem_name] = component_i

            component_counts[component_i] = len(component)

        to_add_edge = None
        for u, v, w, s in tqdm(to_filter_edges[::-1], total=len(to_filter_edges), unit=' final edges'):

            u_c = member_to_component[u]
            v_c = member_to_component[v]

            u_c_realm = component_to_realm[u_c]
            v_c_realm = component_to_realm[v_c]

            if u_c_realm == v_c_realm:
                to_add_edge = (u, v, w, s)

            elif pd.notnull(u_c_realm) and pd.notnull(v_c_realm):  # AND
                if u_c_realm != v_c_realm:
                    continue  # DON'T connect
                else:  # TODO In neither case would we be adding edges, should this whole block be skipped?
                    continue

            elif pd.notnull(u_c_realm) or pd.notnull(v_c_realm):  # Not both
                to_add_edge = (u, v, w, s)

            elif pd.isnull(u_c_realm) and pd.isnull(v_c_realm):
                continue  # Don't want to merge two subcomponents that have no realm

            else:
                print('Unknown groups', u, u_c, u_c_realm, v, v_c, v_c_realm)

            if to_add_edge:
                u_e, v_e, w_e, s_e = to_add_edge
                graph.add_edge(u_e, v_e, shared=s_e, weight=w_e)

                # Rebuild dicts
                if u_c != v_c:  # If the components are not the same
                    # Everyone in one of the components now becomes members of the other
                    greater_component = u_c if component_counts[u_c] >= component_counts[v_c] else v_c
                    lesser_component = v_c if component_counts[v_c] < component_counts[u_c] else u_c

                    # == between component_counts does occur, though it is generally a rare occurrence

                    # Update counts
                    component_counts[greater_component] += component_counts[lesser_component]
                    del component_counts[lesser_component]

                    # Remove component that no longer exists
                    # del component_to_realm[lesser_component]  # FUTURE vertices don't know component no longer exists

                    # Update assignments - ONLY need two, only one edge is added
                    member_to_component[u] = greater_component
                    member_to_component[v] = greater_component

                else:  # One of the previous vertex connected two components, and now they're in the same component
                    continue
                    # No need to change memberships

        return

    if enhanced_sensitivity:
        logger.info(f'Enabling enhanced sensitivity')
        iterative_repair_2nd(ugraph, edges_to_resolve)

    # And finally, modify dataframe
    logger.info(f'Modifying annotation data and writing final graph')

    isolates = ugraph.vs.select(_degree_eq=0)
    isolates_indices = [v.index for v in isolates]
    ugraph.vs[isolates_indices]['isolated'] = True

    if not modifying_df.empty:  # TODO Use graph directly, instead of another connected component enumeration
        for i, members in enumerate(ugraph.connected_components(mode='weak')):
            ugraph.vs[members]['component'] = str(i)
            members = ugraph.vs[members]['name']
            modifying_df.loc[members, 'component'] = str(i)

    # ugraph.vs['component'] = ugraph.connected_components(mode='weak').membership

    write_igraphz(ugraph, out_fp)

    return ugraph
